#include <iostream>
#include <vector>
#include <thread>
#include <mutex>

using namespace std;

void incrementCounterUsingMutex(int *pCounter, mutex &counterMutex) {
    for (int i = 0; i < 10000000; i++) {
        counterMutex.lock();
        (*pCounter)++;
        counterMutex.unlock();
    }
}

void incrementCounterUsingLockGuard(int *pCounter, mutex &counterMutex) {
    for (int i = 0; i < 10000000; i++) {
        // Performance check: what if the locking is done outside of the for?
        lock_guard<mutex> guard(counterMutex);
        (*pCounter)++;
    }
}

int main() {
    int counter = 0;
    mutex counterMutex;

    vector<thread> workers;
    for (int workerIndex = 0; workerIndex < 2; workerIndex++) {
        workers.push_back(thread(incrementCounterUsingMutex, &counter, ref(counterMutex)));
    }
    for (int workerIndex = 0; workerIndex < 2; workerIndex++) {
        workers.push_back(thread(incrementCounterUsingLockGuard, &counter, ref(counterMutex)));
    }

    for (auto &worker : workers) {
        worker.join();
    }

    cout << "Value of the counter variable: " << counter << endl;

    return 0;
}

