#include <chrono>
#include <iostream>
#include <vector>
#include <string>
#include <algorithm>
#include <thread>
#include <mutex>

#define WORK_ITERATIONS 1

using namespace std;
using namespace std::chrono;


class Tool {
public:
    string mName;
    int mAvailableUnits;

    Tool(string name, int availableUnits) : mName(name), mAvailableUnits(availableUnits) {}
};

class Guy {
private:
    string mName;
    Tool &mTool1;
    Tool &mTool2;
    mutex &mMutexTool1;
    mutex &mMutexTool2;

    void work() {
        // Heavy work of a guy, just a random delay...
        this_thread::sleep_for(chrono::milliseconds(rand() % 30));
    }

public:
    Guy(string name, Tool &tool1, Tool &tool2, mutex &mutexTool1, mutex &mutexTool2)
        : mName(name), mTool1(tool1), mTool2(tool2), mMutexTool1(mutexTool1), mMutexTool2(mutexTool2)  {}

    void startGuy() {
        // Each guy is iteratively doing the job.
        for (int iter = 0; iter < WORK_ITERATIONS; ++iter) {

            // Work with one tool
            lock_guard<mutex> guardTool1(mMutexTool1);
            cout << "Guy " << mName << " borrowed: " << mTool1.mName << "." << std::endl;
            work();

            // Work with two tools
            lock_guard<mutex> guardTool2(mMutexTool2);
            cout << "Guy " << mName << " borrowed: " << mTool2.mName << "." << std::endl;
            work();

            // Releases both tools.
        }

        return;
    }
};

int main() {
    Tool hammer("hammer", 1);
    Tool screwDriver("screwdriver", 1);
    Tool saw("saw", 1);

    mutex hammerMutex;
    mutex screwDriverMutex;
    mutex sawMutex;

    Guy libor("Libor", hammer, screwDriver, hammerMutex, screwDriverMutex);
    Guy honza("Honza", screwDriver, saw, screwDriverMutex, sawMutex);
    Guy premek("Premek", saw, hammer, sawMutex, hammerMutex);

    vector<thread> guyThreads;
    guyThreads.emplace_back(&Guy::startGuy, libor);
    guyThreads.emplace_back(&Guy::startGuy, honza);
    guyThreads.emplace_back(&Guy::startGuy, premek);

    for (auto &guyThread: guyThreads) {
        guyThread.join();
    }

    return 0;
}
